﻿/*    
Nemulation: A .NET based framework/platform for virtualization applications
Copyright (C) 2011 Bryan Perris

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Reflection.Emit;
using System.Reflection;
using System.Security;
using NEmu.Debugging;
using System.Globalization;

namespace NEmu.Code.Emitting.IL
{
    public abstract class ILEmitterBase : EmitterBase
    {
        private int m_LocalCount;
        private int m_LocalOffset;
        private long m_CurrentAddress;
        private long m_LastAddress;
        private ILGenerator m_CodeGenerator;
        private Dictionary<long, Label> m_Labels;
        private bool m_DumpCode = false;


        /// <summary>
        /// Generates a function compiled in IL code
        /// </summary>
        /// <param name="address">the entry point address of the code to starting parsing from</param>
        /// <param name="codeCoreState">the CPU core state of the code</param>
        /// <returns>a delegate of the compiled function</returns>
        public override CompiledFunction CompileFunction(long address, object codeCoreState)
        {
            // Set up recompiler variables
            InitState();

            // Create the dyanmic method to compile
            DynamicMethod dynamicMethod = new DynamicMethod(
                "EmittedMethod_" + address.ToString("X8", CultureInfo.CurrentCulture),
                typeof(void),
                new Type[] { typeof(object) },
                typeof(ILEmitterBase).Module);

            EmuLog.DebugLog("Emitting " + dynamicMethod.Name);

            // Get the IL emitter instance
            m_CodeGenerator = dynamicMethod.GetILGenerator();

            // Set the current fake PC to the passed in address
            m_CurrentAddress = address;

            // Run the emitter
            RunEmitter(address, codeCoreState);

            if (IsDumpAssemblies)
            {
                AssemblyName assn = new AssemblyName("GeneratedCode_" + address.ToString("X8", CultureInfo.CurrentCulture));
                AssemblyBuilder ab = AppDomain.CurrentDomain.DefineDynamicAssembly(assn, AssemblyBuilderAccess.Save);
                ModuleBuilder mb = ab.DefineDynamicModule(assn.Name, assn.Name + ".dll");
                TypeBuilder tb = mb.DefineType("MyTestType");
                MethodBuilder mtb = tb.DefineMethod("SysFunc_" + address.ToString("x", CultureInfo.CurrentCulture),
                    MethodAttributes.Public | MethodAttributes.Static, CallingConventions.Standard, typeof(void), new Type[] { codeCoreState.GetType() });
                m_CodeGenerator = mtb.GetILGenerator();
                InitState();
                m_CurrentAddress = address;
                RunEmitter(address, codeCoreState);
                tb.CreateType();
                ab.Save(assn.Name + ".dll");
            }

            return (CompiledFunction)dynamicMethod.CreateDelegate(typeof(CompiledFunction));
        }

        [SuppressUnmanagedCodeSecurityAttribute]
        protected abstract void RunEmitter(long address, object coreState);

        #region Protected

        protected void InitState()
        {
            m_LocalCount = 0;
            m_LocalOffset = 0;
            m_CurrentAddress = 0;
            m_LastAddress = 0;
            m_Labels = new Dictionary<long, Label>();
        }

        protected void SaveCurrentAddress()
        {
            m_LastAddress = m_CurrentAddress;
        }

        protected void UpdateLocalOffsetBase()
        {
            m_LocalOffset = m_LocalCount;
        }

        protected void MarkLabel()
        {
            Label lb = m_CodeGenerator.DefineLabel();
            m_CodeGenerator.MarkLabel(lb);
            m_Labels.Add(m_LastAddress, lb);
        }

        protected Label GetLabel(ushort address)
        {
            Label lb;
            m_Labels.TryGetValue(address, out lb);
            return lb;
        }

        protected int GetResolvedLocal(int offset)
        {
            return (m_LocalOffset + offset);
        }

        #endregion

        #region Common Code Generation

        protected void EmitLocal(Type type, bool pinned)
        {
            m_LocalCount++;
            m_CodeGenerator.DeclareLocal(type, pinned);
        }

        protected void EmitUInt16Constant(ushort value)
        {
            m_CodeGenerator.Emit(OpCodes.Ldc_I4, value);
            m_CodeGenerator.Emit(OpCodes.Conv_U2);
        }

        protected void EmitByteConstant(byte value)
        {
            m_CodeGenerator.Emit(OpCodes.Ldc_I4_S, value);
            m_CodeGenerator.Emit(OpCodes.Conv_U1);
        }

        protected void EmitIntConstant(int value)
        {
            switch (value)
            {
                case 0: m_CodeGenerator.Emit(OpCodes.Ldc_I4_0); break;
                case 1: m_CodeGenerator.Emit(OpCodes.Ldc_I4_1); break;
                case 2: m_CodeGenerator.Emit(OpCodes.Ldc_I4_2); break;
                case 3: m_CodeGenerator.Emit(OpCodes.Ldc_I4_3); break;
                case 4: m_CodeGenerator.Emit(OpCodes.Ldc_I4_4); break;
                case 5: m_CodeGenerator.Emit(OpCodes.Ldc_I4_5); break;
                case 6: m_CodeGenerator.Emit(OpCodes.Ldc_I4_6); break;
                case 7: m_CodeGenerator.Emit(OpCodes.Ldc_I4_7); break;
                case 8: m_CodeGenerator.Emit(OpCodes.Ldc_I4_8); break;
                default: m_CodeGenerator.Emit(OpCodes.Ldc_I4_S, value); break;
            }
        }

        protected void EmitLowByte()
        {
            m_CodeGenerator.Emit(OpCodes.Ldc_I4_S, (ushort)0xFF);
            m_CodeGenerator.Emit(OpCodes.And);
            m_CodeGenerator.Emit(OpCodes.Conv_U1);
        }

        protected void EmitHighByte()
        {
            m_CodeGenerator.Emit(OpCodes.Ldc_I4_S, (ushort)0xFF);
            m_CodeGenerator.Emit(OpCodes.Conv_I2);
            m_CodeGenerator.Emit(OpCodes.And);
            EmitByteConstant(8);
            m_CodeGenerator.Emit(OpCodes.Shr_Un);
            m_CodeGenerator.Emit(OpCodes.Conv_U1);
        }

        protected void EmitLocalStore(int localOffset)
        {
            ILGenerator.Emit(OpCodes.Stloc_S, GetResolvedLocal(localOffset));
        }

        protected void EmitLocalLoad(int localOffset)
        {
            ILGenerator.Emit(OpCodes.Ldloc_S, GetResolvedLocal(localOffset));
        }

        protected void EmitNop()
        {
            m_CodeGenerator.Emit(OpCodes.Nop);
        }

        protected void EmitCoreStateObject()
        {
            m_CodeGenerator.Emit(OpCodes.Ldarg_0);
        }

        #endregion

        #region Properties

        /// <summary>
        /// Whether should each generated method should be dumped into an assembly
        /// </summary>
        public bool IsDumpAssemblies
        {
            get { return m_DumpCode; }
            set { m_DumpCode = true; }
        }

        /// <summary>
        /// The current instance of the IL generator being used for recompilation
        /// </summary>
        public ILGenerator ILGenerator
        {
            get { return m_CodeGenerator; }
        }

        protected long CurrentAddress
        {
            get { return m_CurrentAddress; }
            set { m_CurrentAddress = value; }
        }

        protected long LastAddress
        {
            get { return m_LastAddress; }
        }

        #endregion
    }
}
