﻿/*    
Nemulation: A .NET based framework/platform for virtualization applications
Copyright (C) 2011 Bryan Perris

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime.Serialization;

namespace NEmu.Memory
{
    [Serializable]
    public abstract class MemoryUnit : IDisposable
    {
        private long m_MemBound;
        private Dictionary<long, MemoryPage> m_Pages;
        private MemoryPage m_CurrentPage;
        private long m_CurrentAddressOffset = 0;
        private long m_CurrentAddressBound = 0;
        private VirtualMachine m_ParentMachine;
        private bool disposed;
        private bool m_Initialized;

        protected MemoryUnit(VirtualMachine machine)
        {
            m_Initialized = false;
            m_ParentMachine = machine;
            m_MemBound = 0;
            m_CurrentAddressOffset = 0;
            m_CurrentAddressBound = 0;
            m_Pages = new Dictionary<long, MemoryPage>();
            m_CurrentPage = null;
        }

        ~MemoryUnit()
        {
            Dispose(false);
        }

        public void Initialize()
        {
            AllocatePages();
            m_Initialized = true;
        }

        protected abstract void AllocatePages();

        public virtual void Write(long address, byte value)
        {
            lock (this)
            {
                FindPage(address).Write((int)(address - m_CurrentAddressOffset), value);
            }
        }

        public virtual int Read(long address)
        {
            lock (this)
            {
                return FindPage(address).Read((int)(address - m_CurrentAddressOffset));
            }
        }

        public void Reset()
        {
            foreach (MemoryPage page in m_Pages.Values)
            {
                page.Clear();
            }
        }

        public virtual long Size
        {
            get { return m_MemBound; }
        }

        public virtual byte this[long index]
        {
            get
            {
                return (byte)Read(index);
            }
            set
            {
                Write(index, value);
            }
        }

        public MemoryPage FindPage(long address)
        {
            if (!m_Initialized)
                throw new InvalidOperationException("Memory has not been initialized");

            // If the address isn't out of bounds, use the same page
            if (address >= m_CurrentAddressOffset && address < m_CurrentAddressBound)
            {
                return m_CurrentPage;
            }

            // Else we need to figure out the new page to access
            MemoryPage page = null;

            // Check each offset address and see where address falls in range with
            foreach (long offset in m_Pages.Keys)
            {
                // if address is >= of that key, it is a possible page
                if (address >= offset)
                {
                    // Get the reference of the page
                    if (m_Pages.TryGetValue(offset, out page))
                    {
                        // if the address within bounds of the page, then get it, else keep searching
                        if (address < (offset + page.Size))
                        {
                            m_CurrentPage = page;
                            m_CurrentAddressBound = offset + page.Size;
                            m_CurrentAddressOffset = offset;
                            return page;
                        }
                    }
                }
            }

            return page;
        }

        protected void AddPage(MemoryPage page)
        {
            m_Pages.Add(m_MemBound, page);
            m_MemBound += (page.Size + 1);
        }

        protected void AddManualPage(long address, MemoryPage page)
        {
            m_Pages.Add(address, page);
        }

        protected void ClearPages()
        {
            m_Pages.Clear();
        }

        public VirtualMachine ParentMachine
        {
            get { return this.m_ParentMachine; }
        }

        public bool IsInitialized
        {
            get { return m_Initialized; }
        }

        #region IDisposable Members

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!disposed)
            {
                if (disposing)
                {
                    foreach (MemoryPage page in m_Pages.Values)
                    {
                        m_Initialized = false;
                        page.Dispose();
                    }
                }

                disposed = true;
            }
        }

        #endregion
    }
}
